//! AI engine.

use crate::{eval, find};
use alloc::vec::Vec;
use core::{future::Future, pin::Pin, task::Poll};
use mino::srs::{Piece, PieceType};
use mino::{Mat, MatBuf};

#[derive(Debug)]
#[non_exhaustive]
pub struct Ai {
    root_matrix: MatBuf,
    root_previews: Vec<PieceType>,
    path: Vec<Piece>,
    cur_mat: MatBuf,
}

fn evaluate(mat: &Mat, depth: usize) -> i32 {
    let w_height = 5;
    let w_ideps = 10;
    let w_mdse = 10;
    let w_pc = 10;

    let mut rating = 0;
    rating += eval::max_height(mat) * w_height;
    rating += eval::i_deps(mat) * w_ideps;
    rating += eval::mystery_mdse(mat) * w_mdse;
    rating += (depth as i32) * w_pc;
    rating
}

impl Ai {
    // TODO: personality config
    pub fn new() -> Self {
        Self {
            root_matrix: MatBuf::new(),
            root_previews: Vec::with_capacity(8),
            path: Vec::with_capacity(8),
            cur_mat: MatBuf::new(),
        }
    }

    pub fn start(
        &mut self,
        init_mat: &Mat,
        init_previews: &[PieceType],
        init_hold: Option<PieceType>,
    ) {
        // init root node
        self.root_matrix.copy_from(init_mat);
        self.root_previews.clear();
        self.root_previews.extend_from_slice(init_previews);
        self.root_previews.extend(init_hold); // TODO: actual hold logic

        // init search state
        self.path.clear();
        self.cur_mat.copy_from(&self.root_matrix);
    }

    pub fn think_1_cycle(&mut self) -> bool {
        let depth = self.path.len();
        if depth >= self.root_previews.len() {
            return true;
        }

        let ty = self.root_previews[depth];
        let locs = find::find_locations(&self.cur_mat, ty, find::cap::All);
        let pcs = locs.map(|loc| Piece { ty, loc });

        let mut rate_mat: MatBuf = MatBuf::new();
        let mut best_mat: MatBuf = MatBuf::new();
        let mut best: Option<(i32, Piece)> = None;

        for pc in pcs {
            rate_mat.copy_from(&self.cur_mat);
            pc.cells().fill(&mut rate_mat);
            rate_mat.clear_lines();
            let rating = evaluate(&rate_mat, depth);

            let best_rating = best.clone().map_or(i32::MAX, |(r, _)| r);
            if rating < best_rating {
                best = Some((rating, pc));
                best_mat.copy_from(&rate_mat);
            }
        }

        let pc = match best {
            Some((_, pc)) => pc,
            None => return true, // no locations; game over
        };

        self.path.push(pc);
        self.cur_mat.copy_from(&best_mat);
        false
    }

    pub fn suggestion(&self) -> Vec<Piece> {
        self.path.clone()
    }
}

pub struct Think<'a>(&'a mut Ai);

impl Future for Think<'_> {
    type Output = ();

    fn poll(mut self: Pin<&mut Self>, _cx: &mut core::task::Context<'_>) -> Poll<Self::Output> {
        let ai: &mut Ai = &mut self.0;

        if ai.think_1_cycle() {
            // TODO: if <limits reached> then return Poll::Ready)
            Poll::Pending
        } else {
            Poll::Ready(())
        }
    }
}
