use anyhow::{anyhow, Context as _};

/// Headless fish simulator.
#[derive(Debug, clap::Parser)]
struct Args {
    #[command(subcommand)]
    cmd: Cmd,
}

#[derive(Debug, clap::Subcommand)]
enum Cmd {
    /// List all reachable moves from the input game state.
    List(ListCmd),
    /// Prints the best suggested play for the input game state.
    Best(BestCmd),
    /// Simulate solo play until some end condition is met.
    Play(PlayCmd),
}

#[derive(Debug, clap::Args)]
struct ListCmd {}

#[derive(Debug, clap::Args)]
struct BestCmd {
    #[command(flatten)]
    settings: Settings,
}

#[derive(Debug, clap::Args)]
struct PlayCmd {
    /// Game seed for generating pieces and garbage.
    #[arg(short = 's', long)]
    seed: u64,
    /// Number of moves to play before exiting. If omitted, will play until top out or
    /// until a dig race is finished.
    #[arg(short = 'n', long)]
    count: Option<u64>,
    /// Plays dig race (AKA "cheese race"), inserting rows of garbage onto the playfield
    /// until the specified number of them have been cleared.
    #[arg(short = 'd', long, name = "ROWS")]
    dig: Option<u64>,
    /// Minimum number of garbage rows present at once during a dig race.
    #[arg(long, name = "MIN-ROWS", default_value = "3")]
    dig_min: u64,
    /// Maximum number of garbage rows present at once during a dig race.
    #[arg(long, name = "MAX-ROWS", default_value = "9")]
    dig_max: u64,
    /// Number of previews available to the bot.
    #[arg(short = 'p', long = "previews", name = "PIECES", default_value = "5")]
    previews: u64,
    #[command(flatten)]
    settings: Settings,
}

#[derive(Debug, clap::Args)]
struct Settings {
    /// Limit analysis by given number of expanded nodes.
    #[arg(short = 'N', long = "limit-nodes")]
    nodes: Option<u64>,
    /// Limit analysis by given number of search steps.
    #[arg(short = 'S', long = "limit-steps")]
    steps: Option<u64>,
    /// Limit analysis to given time in milliseconds.
    #[arg(short = 'T', long = "limit-time")]
    ms: Option<u64>,
    // TODO: personality / heuristics configuration
    // TODO: capabilities
    // TODO: output progress format
}

fn main() -> anyhow::Result<()> {
    let args = <Args as clap::Parser>::parse();
    match args.cmd {
        Cmd::List(_) => list_moves(),
        Cmd::Best(cmd) => print_best_move(cmd.settings),
        Cmd::Play(_) => Err(anyhow!("'play' command is not implemented")),
    }
}

fn list_moves() -> anyhow::Result<()> {
    let input: fish::io::InputState =
        serde_json::from_reader(std::io::stdin()).context("error parsing input state")?;

    let mut output = fish::io::OutputMoves::default();

    let mut mat = input.matrix.to_mat();
    mat.clear_lines();

    let front = input.queue.previews.get(0).copied();
    let hold = input.queue.hold;
    for ty in [front, hold].into_iter().flatten() {
        for loc in fish::find_locations(&mat, ty, fish::find::cap::All) {
            output.moves.push(fish::io::OutputMove {
                location: mino::Piece { ty, loc }.into(),
                spin: fish::io::Spin::None,
            });
        }
    }

    serde_json::to_writer(std::io::stdout(), &output).context("error serializing output")?;
    println!();

    Ok(())
}

fn print_best_move(_settings: Settings) -> anyhow::Result<()> {
    let input: fish::io::InputState =
        serde_json::from_reader(std::io::stdin()).context("error parsing input state")?;

    let mut mat = input.matrix.to_mat();
    mat.clear_lines();

    // TODO: ai init config, e.g. personality
    let mut ai = fish::Ai::new();

    // TODO: hold
    // TODO: attack state
    ai.start(&mat, &input.queue.previews, input.queue.hold);

    // TODO: resource limits (cycles,nodes,time)
    loop {
        if ai.think_1_cycle() {
            break;
        }
        // TODO: proper logging
        // eprintln!("thinking...");
    }

    // print suggestions trace
    let mut output = fish::io::OutputMoves::default();
    for pc in ai.suggestion() {
        output.moves.push(fish::io::OutputMove {
            location: pc.into(),
            // TODO: spin suggestions
            spin: fish::io::Spin::None,
        });
    }

    serde_json::to_writer(std::io::stdout(), &output).context("error writing output state")?;
    println!();
    Ok(())
}
